@@ -113,14 +113,10 @@ doc_comment::doctest!("../README.md");
 extern crate proc_macro;
 
 use proc_macro::TokenStream;
-use proc_macro_error::{emit_call_site_error, emit_error, proc_macro_error};
+use proc_macro_error::{emit_call_site_error, proc_macro_error};
 use syn::parse_macro_input;
-use syn::Item;
 
-mod autoimpl;
-mod default;
-pub(crate) mod generics;
-mod scope;
+use impl_tools_lib::{autoimpl, AttrImplDefault, ImplDefault, Scope, ScopeAttr};
 
 /// Implement `Default`
 ///
@@ -175,38 +171,16 @@ mod scope;
 #[proc_macro_attribute]
 #[proc_macro_error]
 pub fn impl_default(attr: TokenStream, item: TokenStream) -> TokenStream {
-    let attr = parse_macro_input!(attr as default::Attr);
-    let attr_span = attr.span;
-    if attr.expr.is_some() {
-        let mut toks = item.clone();
-        match parse_macro_input!(item as Item) {
-            Item::Enum(syn::ItemEnum {
-                ident, generics, ..
-            })
-            | Item::Struct(syn::ItemStruct {
-                ident, generics, ..
-            })
-            | Item::Type(syn::ItemType {
-                ident, generics, ..
-            })
-            | Item::Union(syn::ItemUnion {
-                ident, generics, ..
-            }) => {
-                let impl_: TokenStream = attr.gen_expr(&ident, &generics).into();
-                toks.extend(std::iter::once(impl_));
-            }
-            item => {
-                emit_error!(
-                    item,
-                    "default: only supports enum, struct, type alias and union items"
-                );
-            }
-        };
-        toks
-    } else {
-        emit_error!(attr_span, "invalid use outside of `impl_scope!` macro");
-        item
+    let mut toks = item.clone();
+    match syn::parse::<ImplDefault>(attr) {
+        Ok(attr) => toks.extend(TokenStream::from(attr.expand(item.into()))),
+        Err(err) => {
+            emit_call_site_error!(err);
+            // Since this form of invocation only adds implementations, we can
+            // safely output the original item, thus reducing secondary errors.
+        }
     }
+    toks
 }
 
 /// A variant of the standard `derive` macro
@@ -337,17 +311,18 @@ pub fn impl_default(attr: TokenStream, item: TokenStream) -> TokenStream {
 #[proc_macro_error]
 pub fn autoimpl(attr: TokenStream, item: TokenStream) -> TokenStream {
     let mut toks = item.clone();
-    match syn::parse(attr) {
-        Ok(attr) => {
-            let item = parse_macro_input!(item as Item);
-            toks.extend(TokenStream::from(match item {
-                Item::Struct(item) => autoimpl::autoimpl_struct(attr, item),
-                Item::Trait(item) => autoimpl::autoimpl_trait(attr, item),
-                item => {
-                    emit_error!(item, "autoimpl: only supports struct and trait items");
-                    return toks;
-                }
-            }));
+    match syn::parse::<autoimpl::Attr>(attr) {
+        Ok(autoimpl::Attr::ForDeref(ai)) => toks.extend(TokenStream::from(ai.expand(item.into()))),
+        Ok(autoimpl::Attr::ImplTraits(ai)) => {
+            // We could use lazy_static to construct a HashMap for fast lookups,
+            // but given the small number of impls a "linear map" is fine.
+            let find_impl = |path: &syn::Path| {
+                autoimpl::STD_IMPLS
+                    .iter()
+                    .cloned()
+                    .find(|impl_| impl_.path().matches_ident_or_path(path))
+            };
+            toks.extend(TokenStream::from(ai.expand(item.into(), find_impl)))
         }
         Err(err) => {
             emit_call_site_error!(err);
@@ -418,8 +393,12 @@ pub fn autoimpl(attr: TokenStream, item: TokenStream) -> TokenStream {
 #[proc_macro_error]
 #[proc_macro]
 pub fn impl_scope(input: TokenStream) -> TokenStream {
-    let input = parse_macro_input!(input as scope::Scope);
-    scope::scope(input)
-        .unwrap_or_else(|err| err.to_compile_error())
-        .into()
+    let mut scope = parse_macro_input!(input as Scope);
+    scope.apply_attrs(|path| {
+        AttrImplDefault
+            .path()
+            .matches(path)
+            .then(|| &AttrImplDefault as &dyn ScopeAttr)
+    });
+    scope.expand().into()
 }
